from .Converter import BDX
from json import loads, dumps
from io import BytesIO


def ReadBDXFile(path: str) -> BDX:
    """
    Convert BDX file into class BDX
    """
    with open(path, "r+b") as file:
        fileContext: bytes = file.read()
    # get the content of this bdx file
    result = BDX()
    # request a new class object
    result.UnMarshal(fileContext)
    # unmarshal
    return result
    # return


def DumpStructs(BDXObj: BDX, outputPath: str) -> None:
    """
    Convert BDXObj:BDX into bytes and write it into a bdx file(outputPath:str)
    """
    writer: BytesIO = BytesIO(b'')
    # request a new writer
    BDXObj.Marshal(writer)
    # marshal
    with open(outputPath, 'w+b') as file:
        file.write(writer.getvalue())
    # write bytes into a bdx file


def VisualStructs(BDXObj: BDX, outputPath: str) -> None:
    """
    Convert BDXObj:BDX into json data and write it into outputPath:str
    """
    with open(outputPath, 'w+', encoding='utf-8') as file:
        file.write(
            dumps(
                BDXObj.Dumps(),
                sort_keys=True,
                indent=4,
                separators=(', ', ': '),
                ensure_ascii=False
            ))
    # write json datas


def ConvertJSONFileIntoStructs(path: str) -> BDX:
    """
    Read json datas from path:str and convert it into class BDX
    """
    with open(path, 'r+', encoding='utf-8') as file:
        fileContext: str = file.read()
    jsonDatas: dict = loads(fileContext)
    # load json datas from file
    result = BDX()
    # request a new class object
    result.Loads(jsonDatas)
    # loads
    return result
    # return



from io import BytesIO
from struct import pack
import nbtlib

endian: str = '<'
endianWord: str = 'little'


def getValueType(value) -> bytes:
    if type(value) == nbtlib.tag.Byte:
        return b'\x01'
    elif type(value) == nbtlib.tag.Short:
        return b'\x02'
    elif type(value) == nbtlib.tag.Int:
        return b'\x03'
    elif type(value) == nbtlib.tag.Long:
        return b'\x04'
    elif type(value) == nbtlib.tag.Float:
        return b'\x05'
    elif type(value) == nbtlib.tag.Double:
        return b'\x06'
    elif type(value) == nbtlib.tag.ByteArray:
        return b'\x07'
    elif type(value) == nbtlib.tag.String:
        return b'\x08'
    elif type(value) == nbtlib.tag.Compound:
        return b'\x0a'
    elif type(value) == nbtlib.tag.IntArray:
        return b'\x0b'
    elif type(value) == nbtlib.tag.LongArray:
        return b'\x0c'
    else:
        return b'\x09'


def marshalToName(writer: BytesIO, name: str) -> None:
    encodeResult = name.encode(encoding='utf-8')
    writer.write(pack(f'{endian}H', len(encodeResult)) + encodeResult)


def marshalToValue(writer: BytesIO, value, valueType: int) -> None:
    if valueType == 1:
        writer.write(value.to_bytes(
            length=1, byteorder=endianWord, signed=True))
    elif valueType == 2:
        writer.write(pack(f'{endian}h', value))
    elif valueType == 3:
        writer.write(pack(f'{endian}i', value))
    elif valueType == 4:
        writer.write(pack(f'{endian}q', value))
    elif valueType == 5:
        writer.write(pack(f'{endian}f', value))
    elif valueType == 6:
        writer.write(pack(f'{endian}d', value))
    elif valueType == 7 or valueType == 11 or valueType == 12:
        marshalToArray(writer, value, valueType)
    elif valueType == 8:
        marshalToName(writer, str(value))
    elif valueType == 9:
        marshalToList(writer, value)
    elif valueType == 10:
        marshalToCompound(writer, value)


def marshalToArray(writer: BytesIO,value,valueType: int) -> None:
    writer.write(pack(f'{endian}i', len(value)))
    if valueType == 7:
        writer.write(b''.join([pack(f'{endian}b', i) for i in value]))
    elif valueType == 11:
        writer.write(b''.join([pack(f'{endian}i', i) for i in value]))
    elif valueType == 12:
        writer.write(b''.join([pack(f'{endian}q', i) for i in value]))
    else:
        raise TypeError


def marshalToList(writer: BytesIO, value) -> None:
    if len(value) > 0:
        subValueType = getValueType(value[0])
        writer.write(subValueType+pack(f'{endian}i', len(value)))
        for i in value:
            marshalToValue(writer, i, subValueType[0])
    else:
        writer.write(b'\x00' * 5)


def marshalToCompound(writer: BytesIO, value: nbtlib.tag.Compound) -> None:
    for i in value:
        valueType = getValueType(value[i])
        writer.write(valueType)
        marshalToName(writer, i)
        marshalToValue(writer, value[i], valueType[0])
    writer.write(b'\x00')


def MarshalPythonNBTObjectToWriter(writer: BytesIO, value, name: str) -> None:
    valueType = getValueType(value)
    writer.write(valueType)
    marshalToName(writer, name)
    marshalToValue(writer, value, valueType[0])



from .GeneralClass import GeneralClass
from .Operation import *


def GetBDXCommandPool() -> dict[int, GeneralClass]:
    """
    Return dict[commandId(..operationId):int, PythonClass:GeneralClass]
    """
    return {
        1: CreateConstantString(),
        5: PlaceBlockWithBlockStates(),
        6: AddInt16ZValue0(),
        7: PlaceBlock(),
        8: AddZValue0(),
        9: NOP(),
        12: AddInt32ZValue0(),
        13: PlaceBlockWithBlockStatesDeprecated(),
        14: AddXValue(),
        15: SubtractXValue(),
        16: AddYValue(),
        17: SubtractYValue(),
        18: AddZValue(),
        19: SubtractZValue(),
        20: AddInt16XValue(),
        21: AddInt32XValue(),
        22: AddInt16YValue(),
        23: AddInt32YValue(),
        24: AddInt16ZValue(),
        25: AddInt32ZValue(),
        26: SetCommandBlockData(),
        27: PlaceBlockWithCommandBlockData(),
        28: AddInt8XValue(),
        29: AddInt8YValue(),
        30: AddInt8ZValue(),
        31: UseRuntimeIDPool(),
        32: PlaceRuntimeBlock(),
        33: PlaceBlockWithRuntimeId(),
        34: PlaceRuntimeBlockWithCommandBlockData(),
        35: PlaceRuntimeBlockWithCommandBlockDataAndUint32RuntimeID(),
        36: PlaceCommandBlockWithCommandBlockData(),
        37: PlaceRuntimeBlockWithChestData(),
        38: PlaceRuntimeBlockWithChestDataAndUint32RuntimeID(),
        39: AssignDebugData(),
        40: PlaceBlockWithChestData(),
        41: PlaceBlockWithNBTData()
    }



from io import BytesIO
from struct import unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class AddInt8YValue(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'AddInt8YValue'
        self.operationNumber: int = 29
        self.value: int = 0

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(self.value.to_bytes(
            length=1, byteorder='big', signed=True))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.value = unpack('>b', getByte(buffer, 1))[0]



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class AssignDebugData(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'AssignDebugData'
        self.operationNumber: int = 39
        self.length: int = 0
        self.buffer: bytes = b''

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>I', self.length) + self.buffer)

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.length = unpack('>I', getByte(buffer, 4))[0]
        self.buffer = getByte(buffer, self.length)

    def Loads(self, jsonDict: dict) -> None:
        if 'operationDatas' in jsonDict:
            jsonDict = jsonDict['operationDatas']
            self.length = jsonDict['length'] if 'length' in jsonDict else 0
            self.buffer = b''.join(
                [
                    i.to_bytes(length=1, byteorder='big', signed=False)
                    for i in jsonDict['buffer']
                ]
            ) if 'buffer' in jsonDict else b''

    def Dumps(self) -> dict:
        return {
            'operationName': self.operationName,
            'operationNumber': self.operationNumber,
            'operationDatas': {
                'length': self.length,
                'buffer': [i for i in self.buffer]
            }
        }



from ..General.GeneralClass import GeneralClass


class SubtractZValue(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'SubtractZValue'
        self.operationNumber: int = 19



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getString import getByte, getString


class PlaceCommandBlockWithCommandBlockData(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'PlaceCommandBlockWithCommandBlockData'
        self.operationNumber: int = 36
        self.data: int = 0
        self.mode: int = 0
        self.command: str = ''
        self.customName: str = ''
        self.lastOutput: str = ''
        self.tickdelay: int = 0
        self.executeOnFirstTick: bool = True
        self.trackOutput: bool = True
        self.conditional: bool = False
        self.needsRedstone: bool = False

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>H', self.data) + pack('>I', self.mode) + self.command.encode(encoding='utf-8') + b'\x00' + self.customName.encode(encoding='utf-8') + b'\x00' + self.lastOutput.encode(encoding='utf-8') + b'\x00' + pack('>I', self.tickdelay) + self.executeOnFirstTick.to_bytes(
            length=1, byteorder='big', signed=False) + self.trackOutput.to_bytes(length=1, byteorder='big', signed=False) + self.conditional.to_bytes(length=1, byteorder='big', signed=False) + self.needsRedstone.to_bytes(length=1, byteorder='big', signed=False))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.data = unpack('>H', getByte(buffer, 2))[0]
        self.mode = unpack('>I', getByte(buffer, 4))[0]
        self.command = getString(buffer)
        self.customName = getString(buffer)
        self.lastOutput = getString(buffer)
        self.tickdelay = unpack('>I', getByte(buffer, 4))[0]
        self.executeOnFirstTick = bool(getByte(buffer, 1)[0])
        self.trackOutput = bool(getByte(buffer, 1)[0])
        self.conditional = bool(getByte(buffer, 1)[0])
        self.needsRedstone = bool(getByte(buffer, 1)[0])



from ..General.GeneralClass import GeneralClass


class SubtractYValue(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'SubtractYValue'
        self.operationNumber: int = 17



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getString import getByte, getString


class ChestSlot(GeneralClass):
    """
    Note: `ChestSlot` is not a operation
    """

    def __init__(self) -> None:
        self.itemName: str = ''
        self.count: int = 0
        self.data: int = 0
        self.slotID: int = 0

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(self.itemName.encode(encoding='utf-8') + b'\x00' + self.count.to_bytes(length=1, byteorder='big',
                     signed=False) + pack('>H', self.data) + self.slotID.to_bytes(length=1, byteorder='big', signed=False))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.itemName = getString(buffer)
        self.count = getByte(buffer, 1)[0]
        self.data = unpack('>H', getByte(buffer, 2))[0]
        self.slotID = getByte(buffer, 1)[0]


class ChestData(GeneralClass):
    """
    Note: `ChestData` is not a operation
    """

    def __init__(self) -> None:
        self.slotCount: int = 0
        self.chestData: list[ChestSlot] = []

    def Marshal(self, writer: BytesIO) -> None:
        for i in self.chestData:
            i.Marshal(writer)

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.chestData = []
        for _ in range(self.slotCount):
            newSlot = ChestSlot()
            newSlot.UnMarshal(buffer)
            self.chestData.append(newSlot)

    def Loads(self, jsonList: list[dict]) -> None:
        self.slotCount = len(jsonList)
        self.chestData = []
        for i in jsonList:
            newSlot = ChestSlot()
            newSlot.Loads(i)
            self.chestData.append(newSlot)
        self.slotCount = len(self.chestData)

    def Dumps(self) -> list[dict]:
        result: list[dict] = []
        for i in self.chestData:
            result.append(i.Dumps())
        return result



class headerError(Exception):
    """
    header error occurred while reading header in BDX file
    """

    def __init__(self, header: bytes):
        Exception.__init__(self, f'find invalid file header "{header}"')


class readError(Exception):
    """
    BDX file read error
    """

    def __init__(self, errorOccurredPosition: int):
        Exception.__init__(
            self, f'failed to convert this BDX file, and the error occurred at position {errorOccurredPosition}')


class unknownOperationError(Exception):
    """
    Find unknown operation error
    """

    def __init__(self, operationId: int, errorOccurredPosition: int):
        Exception.__init__(
            self, f'an unknown operation {operationId} was found, and the error occurred at position {errorOccurredPosition}')


class signatureError(Exception):
    """
    Error occurred while signing
    """

    def __init__(self, errorContent: str):
        Exception.__init__(self, errorContent)



from io import BytesIO
from struct import pack, unpack
from .structOfChest import ChestData
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class PlaceRuntimeBlockWithChestDataAndUint32RuntimeID(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'PlaceRuntimeBlockWithChestDataAndUint32RuntimeID'
        self.operationNumber: int = 38
        self.runtimeId: int = 0
        self.slotCount: int = 0
        self.data: ChestData = ChestData()

    def Marshal(self, writer: BytesIO) -> None:
        self.data.slotCount = self.slotCount
        writer.write(pack('>I', self.runtimeId) +
                     self.slotCount.to_bytes(length=1, byteorder='big', signed=False))
        self.data.Marshal(writer)

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.runtimeId = unpack('>I', getByte(buffer, 4))[0]
        self.slotCount = getByte(buffer, 1)[0]
        self.data.slotCount = self.slotCount
        self.data.UnMarshal(buffer)

    def Loads(self, jsonDict: dict) -> None:
        if 'operationDatas' in jsonDict:
            jsonDict = jsonDict['operationDatas']
            self.runtimeId = jsonDict['runtimeId'] if 'runtimeId' in jsonDict else 0
            newChestData = ChestData()
            if 'data' in jsonDict:
                newChestData.Loads(jsonDict['data'])
            self.slotCount = self.data.slotCount
            self.data = newChestData

    def Dumps(self) -> dict:
        return {
            'operationName': self.operationName,
            'operationNumber': self.operationNumber,
            'operationDatas': {
                'runtimeId': self.runtimeId,
                'slotCount': len(self.data.chestData),
                'data': self.data.Dumps()
            }
        }



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class PlaceBlock(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'PlaceBlock'
        self.operationNumber: int = 7
        self.blockConstantStringID: int = 0
        self.blockData: int = 0

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>H', self.blockConstantStringID) +
                     pack('>H', self.blockData))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.blockConstantStringID = unpack('>H', getByte(buffer, 2))[0]
        self.blockData = unpack('>H', getByte(buffer, 2))[0]



from .Converter.FileOperation import ReadBDXFile, DumpStructs
from .Converter.FileOperation import VisualStructs, ConvertJSONFileIntoStructs

from .General.GeneralClass import GeneralClass

from .General import Operation
from .General.Pool import GetBDXCommandPool

from .Converter.Signature import Signature
from .Converter.Converter import BDX



import setuptools

with open('requirements.txt', 'r+', encoding='utf-8') as file:
    dependences = file.read().strip().split('\n')

with open('README.md', 'r+', encoding='utf-8') as file:
    long_description = file.read()

with open('version', 'r+', encoding='utf-8') as file:
    version = file.read().strip()

setuptools.setup(
    name='BDXConverter',
    version=version,
    author='Minecraft Muti-Media Organization',
    author_email='TriM-Organization@hotmail.com',
    description='A code library to marshal, unmarshal, visual and reverse visualization of BDX files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/TriM-Organization/BDXConverter',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=dependences,
    python_requires='>=3.9',
)



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class AddInt16XValue(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'AddInt16XValue'
        self.operationNumber: int = 20
        self.value: int = 0

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>h', self.value))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.value = unpack('>h', getByte(buffer, 2))[0]



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class AddInt32ZValue(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'AddInt32ZValue'
        self.operationNumber: int = 25
        self.value: int = 0

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>i', self.value))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.value = unpack('>i', getByte(buffer, 4))[0]
